# Copyright 2025 Foodles (https://www.foodles.co/).
# License AGPL-3.0 or later (https://www.gnu.org/licenses/agpl.html).
import importlib
import inspect
import os
import pathlib
import shutil
import textwrap
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, ForwardRef, Generator, List, Union, get_type_hints
from unittest import mock, TestCase

from odoo_typing_classes_generator.core.generator import (
    Generator as TypingClassesGenerator,
)
from odoo_typing_classes_generator.tests.fake_module_1 import typing as models_typing


class TestGenerator(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.typing_folder_path = Path(
            "odoo_typing_classes_generator/tests/fake_module_1/typing"
        )
        cls.class_file_path = cls.typing_folder_path / "__init__.py"
        cls.stub_file_path = cls.typing_folder_path / "__init__.pyi"

    def _models_file_content(self):
        return textwrap.dedent(
            """\
            class ResCountry:
                pass


            class FakeModel1:
                pass
            """
        )

    def _reset_models_files(self):
        with open(self.class_file_path, "w") as models_file:
            models_file.write(self._models_file_content())
        if self.stub_file_path.exists():
            os.remove(self.stub_file_path)

    def setUp(self):
        self._reset_models_files()

    def tearDown(self):
        self._reset_models_files()

    @contextmanager
    def mock_open(self) -> Generator:
        with mock.patch("builtins.open") as mock_open:
            yield mock_open

    @contextmanager
    def mock_get_addon_package_name(self) -> Generator:
        def _get_addon_package_name(addon_name: str) -> str:
            if not addon_name.startswith("fake_"):
                return f"odoo.addons.{addon_name}"
            return f"odoo_typing_classes_generator.tests.{addon_name}"

        with mock.patch(
            "odoo_typing_classes_generator.core.generator."
            "Generator.get_addon_package_name",
            side_effect=_get_addon_package_name,
        ) as mock_get_addon_package_name:
            yield mock_get_addon_package_name

    @contextmanager
    def mock_get_addon_folder_path(self) -> Generator:
        def _get_addon_folder_path(addon_name: str) -> Path:
            if not addon_name.startswith("fake_"):
                return Path("odoo/addons") / addon_name
            return pathlib.Path(__file__).parent.resolve() / addon_name

        with mock.patch(
            "odoo_typing_classes_generator.core.generator."
            "Generator.get_addon_folder_path",
            side_effect=_get_addon_folder_path,
        ) as mock_get_addon_folder_path:
            yield mock_get_addon_folder_path

    def test_load_models_data(self):
        generator = TypingClassesGenerator(
            addons_path="odoo/addons/typing_classes_generator/tests",
        )
        with self.mock_get_addon_package_name(), self.mock_get_addon_folder_path():
            generator.generate("fake_module_1")
            self.assertTrue(self.class_file_path.exists())
            self.assertTrue(self.stub_file_path.exists())
            with open(self.class_file_path, "r") as models_file:
                models_file_content = models_file.read()
                self.assertEqual(self._models_file_content(), models_file_content)
            shutil.move(str(self.stub_file_path), str(self.class_file_path))
            self._test_models_typing()
        shutil.move(str(self.class_file_path), str(self.stub_file_path))

    def test_load_models_data_generate_all_classes(self):
        generator = TypingClassesGenerator(
            addons_path="odoo/addons/typing_classes_generator/tests",
            generate_all_classes=True,
        )
        with self.mock_get_addon_package_name(), self.mock_get_addon_folder_path():
            generator.generate("fake_module_1")
            self.assertTrue(self.class_file_path.exists())
            self.assertTrue(self.stub_file_path.exists())
            with open(self.class_file_path, "r") as models_file:
                models_file_content = models_file.read()
                self.assertNotEqual(self._models_file_content(), models_file_content)
            shutil.move(str(self.stub_file_path), str(self.class_file_path))
            self._test_models_typing()

    def _test_models_typing(self):
        importlib.reload(models_typing)
        module_member_names = dict(inspect.getmembers(models_typing)).keys()
        self.assertGreater(len(module_member_names), 2)
        self.assertIn("ResCountry", module_member_names)
        self.assertIn("FakeModel1", module_member_names)
        members = dict(models_typing.FakeModel1.__dict__)
        # Testing class declaration
        self.assertEqual(
            members["__orig_bases__"],
            (
                models_typing.Model[
                    Union[ForwardRef("FakeModel1"), ForwardRef("ResPartner")]
                ],
            ),
        )
        self.assertEqual(
            members["__doc__"],
            (
                "\n"
                "    Merged model for fake.model.1, built from:\n"
                "        * odoo_typing_classes_generator.tests.fake_module_1.models.fake_model_1.FakeModel1\n"
                "\n"
                "    And inherits from:\n"
                "        * res.partner\n"
                "    "
            ),
        )

        # Testing fields specific to fake.model.1
        member_annotations = members["__annotations__"]
        self.assertIn("_name", members)
        self.assertEqual(members["_name"], "fake.model.1")
        self.assertIn("a_text_field", member_annotations)
        self.assertEqual(member_annotations["a_text_field"], str)
        self.assertIn("a_related_field", member_annotations)
        self.assertEqual(member_annotations["a_related_field"], str)
        self.assertIn("a_many_to_one_field", member_annotations)
        self.assertEqual(
            member_annotations["a_many_to_one_field"],
            Union[ForwardRef("ResCompany"), bool],
        )
        self.assertIn("a_class_method", members)

        # Testing a_documented_function
        a_documented_function = members["a_documented_function"]
        self.assertNotIsInstance(a_documented_function, classmethod)
        self.assertNotIsInstance(a_documented_function, staticmethod)
        a_documented_function_hints = get_type_hints(a_documented_function)
        a_documented_function_signature = inspect.signature(a_documented_function)
        self.assertEqual(
            list(a_documented_function_signature.parameters.keys()),
            ["self", "_an_argument"],
        )
        self.assertNotIn("self", a_documented_function_hints)
        self.assertEqual(
            a_documented_function_signature.parameters["self"].default,
            inspect.Parameter.empty,
        )
        self.assertIn("_an_argument", a_documented_function_hints)
        self.assertEqual(a_documented_function_hints["_an_argument"], int)
        self.assertEqual(
            a_documented_function_signature.parameters["_an_argument"].default, 1
        )
        self.assertIn("return", a_documented_function_hints)
        self.assertEqual(
            a_documented_function_hints["return"], models_typing.ResCountry
        )

        # Testing an_undocumented_method
        an_undocumented_method = members["an_undocumented_method"]
        self.assertNotIsInstance(an_undocumented_method, classmethod)
        self.assertNotIsInstance(an_undocumented_method, staticmethod)
        an_undocumented_method_signature = inspect.signature(an_undocumented_method)
        self.assertEqual(
            list(an_undocumented_method_signature.parameters.keys()),
            ["self", "_an_argument"],
        )
        self.assertEqual(
            an_undocumented_method_signature.parameters["self"].annotation,
            inspect.Parameter.empty,
        )
        self.assertEqual(
            an_undocumented_method_signature.parameters["self"].default,
            inspect.Parameter.empty,
        )
        self.assertEqual(
            an_undocumented_method_signature.parameters["_an_argument"].annotation,
            inspect.Parameter.empty,
        )
        self.assertEqual(
            an_undocumented_method_signature.parameters["_an_argument"].default, ""
        )
        self.assertEqual(
            an_undocumented_method_signature.return_annotation,
            inspect.Parameter.empty,
        )

        # Testing a_static_method_with_default_arguments
        a_static_method_with_default_arguments = members[
            "a_static_method_with_default_arguments"
        ]
        self.assertIsInstance(a_static_method_with_default_arguments, staticmethod)
        a_static_method_with_default_arguments_signature = inspect.signature(
            a_static_method_with_default_arguments.__func__
        )
        self.assertEqual(
            list(a_static_method_with_default_arguments_signature.parameters.keys()),
            ["an_argument", "another_argument"],
        )
        self.assertEqual(
            a_static_method_with_default_arguments_signature.parameters[
                "an_argument"
            ].annotation,
            Callable[[], Any],
        )
        self.assertEqual(
            a_static_method_with_default_arguments_signature.parameters[
                "an_argument"
            ].default,
            None,
        )
        self.assertEqual(
            a_static_method_with_default_arguments_signature.parameters[
                "another_argument"
            ].annotation,
            inspect.Parameter.empty,
        )
        self.assertEqual(
            a_static_method_with_default_arguments_signature.parameters[
                "another_argument"
            ].default,
            "world",
        )
        self.assertEqual(
            a_static_method_with_default_arguments_signature.return_annotation,
            inspect.Parameter.empty,
        )

        # Testing a_static_method
        a_static_method = members["a_static_method"]
        self.assertIsInstance(a_static_method, staticmethod)
        a_static_method_hints = get_type_hints(a_static_method.__func__)
        a_static_method_signature = inspect.signature(a_static_method.__func__)
        self.assertEqual(
            list(a_static_method_signature.parameters.keys()), ["an_argument"]
        )
        self.assertIn("an_argument", a_static_method_hints)
        self.assertEqual(a_static_method_hints["an_argument"], bool)
        self.assertEqual(
            a_static_method_signature.parameters["an_argument"].default,
            inspect.Parameter.empty,
        )
        self.assertIn("return", a_static_method_hints)
        self.assertEqual(a_static_method_hints["return"], int)

        # Testing a_class_method
        a_class_method = members["a_class_method"]
        self.assertIsInstance(a_class_method, classmethod)
        a_class_method_hints = get_type_hints(a_class_method.__func__)
        a_class_method_signature = inspect.signature(a_class_method.__func__)
        self.assertEqual(
            list(a_class_method_signature.parameters.keys()), ["cls", "an_argument"]
        )
        self.assertNotIn("cls", a_class_method_hints)
        self.assertEqual(
            a_class_method_signature.parameters["cls"].default,
            inspect.Parameter.empty,
        )
        self.assertIn("an_argument", a_class_method_hints)
        self.assertEqual(a_class_method_hints["an_argument"], str)
        self.assertEqual(
            a_class_method_signature.parameters["an_argument"].default,
            inspect.Parameter.empty,
        )
        self.assertIn("return", a_class_method_hints)
        self.assertEqual(a_class_method_hints["return"], List[str])



import ast
import datetime
import importlib
import inspect
import logging
import textwrap
from collections import defaultdict
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Union,
    get_type_hints,
)

from odoo import fields, models

from .models import (
    AddonData,
    FieldData,
    FunctionArgumentData,
    FunctionData,
    ImportData,
    MergedModelData,
    ModelData,
    TypeData,
)

_logger = logging.getLogger(__name__)


class Generator:
    python_type_by_odoo_field_type = {
        "date": datetime.date,
        "datetime": datetime.datetime,
        "time": datetime.time,
        "binary": bytes,
        "integer": int,
        "char": str,
        "text": str,
        "selection": str,
        "html": str,
        "serialized": str,
        "job_serialized": str,
        "float": float,
        "monetary": float,
        "boolean": bool,
        "many2one_reference": int,
        "reference": str,
    }

    def __init__(self, addons_path: str, generate_all_classes: bool = False):
        self.addons_path = Path(addons_path)
        self.generate_all_classes = generate_all_classes

        self.assets_path = Path(__file__).parent.parent / "assets"

        self.models_data_by_odoo_model_name: Dict[str, List[ModelData]] = defaultdict(
            list
        )
        self.addon_data_by_name: Dict[str, AddonData] = {}
        self.related_fields_to_resolve_by_odoo_model_name: Dict[
            str,
            Set[str],
        ] = defaultdict(set)
        self.modules_to_scan = {"models", "wizards"}
        self.merged_models_data_by_odoo_model_name: Dict[str, MergedModelData] = {}

    def generate(self, addon_name: str):
        # List of models.Model defined in the addon
        _addon_data = self._load_models_data(addon_name)
        # Create the merged ModelData for each Odoo model
        self._merge_models_data(addon_name)
        # Resolve the related fields
        self._resolve_related_fields(addon_name)
        # Create the stubs
        self._generate_stubs(addon_name)

    def _load_models_data(self, addon_name: str) -> AddonData:
        # This addon has already been processed, nothing to do
        if addon_name in self.addon_data_by_name:
            return self.addon_data_by_name[addon_name]
        _logger.info(f"[{addon_name}] Loading all models data")
        addon_data = AddonData(
            name=addon_name,
            models_data_by_odoo_model_name=defaultdict(list),
            dependencies=set(),
        )
        self.addon_data_by_name[addon_name] = addon_data

        addon_package_name = self.get_addon_package_name(addon_name)
        addon_module = importlib.import_module(addon_package_name)
        self._load_models_data_in_module(addon_name, addon_data, addon_module)
        if addon_name == "base":
            return addon_data

        manifest_path = Path(addon_module.__file__).parent / "__manifest__.py"
        with open(manifest_path) as manifest_file:
            manifest_content = manifest_file.read()
        manifest = ast.literal_eval(manifest_content)
        addon_dependencies = set(manifest.get("depends", []))
        addon_dependencies.add("base")
        for addon_dependency in addon_dependencies:
            addon_data.dependencies.add(self._load_models_data(addon_dependency))
        return addon_data

    def _load_models_data_in_module(
        self,
        addon_name: str,
        addon_data: AddonData,
        module: Any,
    ):
        for _model_name, model in inspect.getmembers(
            module, Generator.is_member_odoo_model
        ):
            package_name = module.__name__
            odoo_model_name = Generator.get_odoo_model_name(model)
            if not odoo_model_name:
                continue
            model_data = self._get_model_data(
                addon_name,
                package_name,
                odoo_model_name,
                model,
            )
            self.models_data_by_odoo_model_name[odoo_model_name].append(model_data)
            addon_data.models_data_by_odoo_model_name[odoo_model_name].append(
                model_data
            )
        check_submodules = inspect.getfile(module).endswith("/__init__.py")
        for _member_name, member in inspect.getmembers(module, inspect.ismodule):
            try:
                member_file_path = inspect.getfile(member)
            except:  # noqa: S112,E722
                continue
            if (
                str(self.get_addon_folder_path(addon_name)) in member_file_path
                and check_submodules
            ):
                self._load_models_data_in_module(addon_name, addon_data, member)

    def _get_model_data(
        self,
        addon_name: str,
        package_name: str,
        odoo_model_name: str,
        model: type,
    ) -> ModelData:
        _logger.info(
            f"[{addon_name}, {odoo_model_name}] Processing class {model.__class__.__module__}.{model.__name__}"
        )
        model_data = ModelData(
            package_name=package_name,
            odoo_model_name=odoo_model_name,
            source_class_name=model.__name__,
            addon_name=addon_name,
            inherited_model_names=set(Generator.get_inherits(model)),
            field_data_by_name={},
            function_data_by_name={},
            imports=set(),
            abstract=model._abstract,
            transient=model._transient,
        )
        # TODO handle non-Odoo fields and properties
        for field_name, field in inspect.getmembers(
            model,
            Generator.is_member_odoo_field,
        ):
            field_data = self._get_field_data(
                odoo_model_name,
                field_name,
                field,
            )
            if not field_data:
                continue
            if (
                field_data.type
                and not isinstance(field_data.type, str)
                and field_data.type.value
                in (datetime.date, datetime.datetime, datetime.time)
            ):
                model_data.imports.add(
                    ImportData(
                        package_name="datetime",
                        class_name=field_data.type.value.__name__,
                        alias=None,
                    )
                )
            model_data.field_data_by_name[field_name] = field_data
        for function_name, function in inspect.getmembers(
            model, Generator.is_member_function
        ):
            # We want to ignore name-mangled functions
            if not function_name.endswith("__") and "__" in function_name:
                continue
            _logger.debug(f"Processing {model_data.source_class_name}.{function_name}")
            function_data = FunctionData(
                name=function_name,
                args=[],
            )
            hints = get_type_hints(function)
            if "return" in hints:
                function_data.type = TypeData(hints["return"])
                model_data.imports |= function_data.type.imports

            attr = inspect.getattr_static(
                model,
                function_name,
            )
            function_data.is_class_method = isinstance(attr, classmethod)
            function_data.is_static_method = isinstance(attr, staticmethod)
            # Gets the unbound function to get the correct signature
            if hasattr(function, "__func__"):
                signature = inspect.signature(function.__func__)
            else:
                signature = inspect.signature(function)
            for arg_index, (arg_name, arg) in enumerate(signature.parameters.items()):
                if arg.annotation is not inspect.Parameter.empty:
                    function_argument_data = FunctionArgumentData(
                        name=arg_name,
                        type=TypeData(hints.get(arg_name)),
                    )
                    model_data.imports |= function_argument_data.type.imports
                else:
                    function_argument_data = FunctionArgumentData(
                        name=arg_name,
                        type=None,
                    )
                if inspect.isfunction(arg.default):
                    # If the default is a function, we don't know its type
                    function_argument_data.default = None
                    function_argument_data.has_default = True
                    if function_argument_data.type is None:
                        default_signature = inspect.signature(arg.default)
                        default_arg_type: List[type] = [
                            Any for _ in default_signature.parameters
                        ]
                        function_argument_data.type = TypeData(
                            Callable[default_arg_type, Any]
                        )
                    model_data.imports |= function_argument_data.type.imports
                elif arg.default is not inspect.Parameter.empty:
                    function_argument_data.default = arg.default
                    function_argument_data.has_default = True
                    default_class = arg.default.__class__
                    model_data.imports |= TypeData.collect_imports(default_class)
                if arg.kind == inspect.Parameter.VAR_POSITIONAL:
                    function_argument_data.is_vararg = True
                elif arg.kind == inspect.Parameter.VAR_KEYWORD:
                    function_argument_data.is_kwarg = True
                function_data.args.append(function_argument_data)
            model_data.function_data_by_name[function_name] = function_data
        return model_data

    def _get_field_data(
        self,
        odoo_model_name: str,
        field_name: str,
        field: fields.Field,
    ) -> Optional[FieldData]:
        field_data = FieldData(name=field_name, type="", related=None)
        field_class = field.__class__.__name__
        if isinstance(field, (fields.Many2one, fields.One2many, fields.Many2many)):
            if "related" in field.args:
                self.related_fields_to_resolve_by_odoo_model_name[odoo_model_name].add(
                    field.args["related"]
                )
                return None
            if "comodel_name" not in field.args:
                return None
            field_data.type = field.args["comodel_name"]
            return field_data
        if field.type not in self.python_type_by_odoo_field_type:
            _logger.warning(
                "Field %s with type %s is not supported",
                field_name,
                field_class,
            )
            return None
        field_data.type = TypeData(self.python_type_by_odoo_field_type[field.type])
        return field_data

    @staticmethod
    def get_addon_package_name(addon_name: str) -> str:
        return f"odoo.addons.{addon_name}"

    def get_addon_folder_path(self, addon_name: str) -> Path:
        return self.addons_path / addon_name

    def _merge_models_data(self, addon_name: str) -> None:
        package_name = f"{self.get_addon_package_name(addon_name)}.typing.models"
        for odoo_model_name, models_data in self.models_data_by_odoo_model_name.items():
            _logger.info(f"[{addon_name}, {odoo_model_name}] Merging all models data")
            merged_model_data = MergedModelData(
                package_name=package_name,
                odoo_model_name=odoo_model_name,
                source_class_name="",
                addon_name=addon_name,
                inherited_model_names={odoo_model_name},
                field_data_by_name={},
                function_data_by_name={},
                imports=set(),
                source_models=[],
                abstract=False,
                transient=False,
            )
            for model_data in sorted(models_data, key=lambda m: m.package_name):
                if model_data.is_base_definition:
                    merged_model_data.source_class_name = model_data.source_class_name
                    merged_model_data.class_name = model_data.class_name
                    merged_model_data.abstract = model_data.abstract
                    merged_model_data.transient = model_data.transient
                merged_model_data.field_data_by_name.update(
                    model_data.field_data_by_name
                )
                merged_model_data.inherited_model_names |= (
                    model_data.inherited_model_names
                )
                for (
                    function_name,
                    function_data,
                ) in model_data.function_data_by_name.items():
                    if function_name not in merged_model_data.function_data_by_name:
                        merged_model_data.function_data_by_name[
                            function_name
                        ] = function_data
                        continue
                    merged_function_data = merged_model_data.function_data_by_name[
                        function_name
                    ]
                    if merged_function_data.type:
                        merged_function_data.type = function_data.type
                    for merged_arg in merged_function_data.args:
                        if merged_arg.type:
                            continue
                        arg = next(
                            (
                                arg
                                for arg in function_data.args
                                if (arg.name == merged_arg.name and arg.type)
                            ),
                            False,
                        )
                        if not arg:
                            continue
                        merged_arg.type = arg.type
                merged_model_data.field_data_by_name.update(
                    model_data.field_data_by_name
                )
                merged_model_data.imports.update(model_data.imports)
                merged_model_data.source_models.append(model_data)
            self.merged_models_data_by_odoo_model_name[
                odoo_model_name
            ] = merged_model_data

    @classmethod
    def get_odoo_model_name(cls, model: Any) -> Optional[str]:
        if model._name:
            return model._name
        inherits = Generator.get_inherits(model)
        if not inherits:
            return None
        return inherits[0]

    @classmethod
    def get_inherits(cls, model: Any) -> List[str]:
        if not hasattr(model, "_inherit"):
            return []
        inherit: Union[str, List[str]] = model._inherit or []
        if isinstance(inherit, str):
            return [inherit]
        return inherit

    @classmethod
    def is_member_odoo_model(cls, member: Any) -> bool:
        return inspect.isclass(member) and issubclass(member, models.AbstractModel)

    @classmethod
    def is_member_odoo_field(cls, member: Any) -> bool:
        return issubclass(member.__class__.__class__, fields.MetaField)

    @classmethod
    def is_member_function(cls, member: Any) -> bool:
        return inspect.isfunction(member) or inspect.ismethod(member)

    def _resolve_related_fields(self, addon_name: str) -> None:
        _logger.info(f"[{addon_name} Resolving all the related fields")
        related_fields_to_resolve_before: Set[str] = set()
        related_fields_to_resolve = {
            f'"{odoo_model_name}".{field_name}'
            for odoo_model_name, field_names in self.related_fields_to_resolve_by_odoo_model_name.items()
            for field_name in field_names
        }
        while related_fields_to_resolve != related_fields_to_resolve_before:
            related_fields_to_resolve_before = set(related_fields_to_resolve)
            for (
                odoo_model_name,
                related_fields,
            ) in self.related_fields_to_resolve_by_odoo_model_name.items():
                model_data = self.merged_models_data_by_odoo_model_name[odoo_model_name]
                for field_name in related_fields:
                    field_key = f'"{odoo_model_name}".{field_name}'
                    if field_key not in related_fields_to_resolve:
                        continue
                    if field_name not in model_data.field_data_by_name:
                        continue
                    field_data = model_data.field_data_by_name[field_name]
                    field_type = self._resolve_related_field(model_data, field_data)
                    if field_type:
                        field_data.field_type = field_type
                        related_fields_to_resolve.remove(
                            f'"{odoo_model_name}".{field_name}'
                        )

    def _resolve_related_field(
        self, model_data: MergedModelData, field_data: FieldData
    ) -> Optional[Union[type, str]]:
        if field_data.type:
            return field_data.type
        related_field_name, related_subfield_name = field_data.related.split(".", 1)
        related_field_data = model_data.field_data_by_name[related_field_name]
        related_model_data = self.merged_models_data_by_odoo_model_name[
            related_field_name
        ]
        while related_subfield_name:
            if not related_field_data.type:
                # This field is also related, we need to resolve it first
                return None
            related_field_data = related_model_data.field_data_by_name[
                related_field_name
            ]
            (
                related_field_name,
                related_subfield_name,
            ) = related_subfield_name.related.split(".", 1)
            related_model_data = self.merged_models_data_by_odoo_model_name[
                related_field_name
            ]
        return related_field_data.type

    def _generate_stubs(self, addon_name: str) -> None:
        _logger.info(f"[{addon_name}] Generating the stubs")
        typing_folder = self.get_addon_folder_path(addon_name) / "typing"
        typing_folder.mkdir(parents=False, exist_ok=True)

        models_path = typing_folder / "__init__.py"
        if self.generate_all_classes:
            with open(models_path, "w") as models_file:
                models_file.write(self._generate_class_file_content())
        elif not models_path.exists():
            with open(models_path, "w") as models_file:
                models_file.write(
                    textwrap.dedent(
                        """\
                    # Add below the classes you want to use for typing,
                    # see the documentation for more information

                    """
                    )
                )

        models_stubs_path = typing_folder / "__init__.pyi"
        with open(models_stubs_path, "w") as models_stubs_file:
            models_stubs_file.write(self._generate_stub_file_content())

    def _generate_class_file_content(self) -> str:
        class_file_content = textwrap.dedent(
            """\
            # This file is auto-generated by Odoo Typing Classes Generator


            """
        )
        all_class_definitions = [
            merged_model_data.get_class_definition()
            for merged_model_data in self.merged_models_data_by_odoo_model_name.values()
        ]
        class_file_content += "\n\n".join(sorted(all_class_definitions))
        return class_file_content

    def _generate_stub_file_content(self) -> str:
        with open(self.assets_path / "template.py", "r") as template_stub_file:
            stub_file_content = template_stub_file.read()
        all_import_lines: Set[str] = set()
        all_stub_definitions: List[str] = []
        model_class_names: Set[str] = set()
        for merged_model_data in self.merged_models_data_by_odoo_model_name.values():
            model_class_names.add(merged_model_data.unique_class_name)
            import_lines, stub_definition = merged_model_data.get_stub_definition(
                self.merged_models_data_by_odoo_model_name
            )
            all_import_lines.update(import_lines)
            all_stub_definitions.append(stub_definition)
        stub_file_content = stub_file_content.replace(
            "# odoo-typing-classes-generator: imports-insertion-point",
            "\n".join(sorted(all_import_lines)) + "\n",
        )
        stub_file_content = stub_file_content.replace(
            "# odoo-typing-classes-generator: classes-insertion-point",
            "\n\n".join(sorted(all_stub_definitions)),
        )
        return stub_file_content



# This file is auto-generated by Odoo Typing Classes Generator

from datetime import datetime
from typing import (
    Any,
    Callable,
    Collection,
    Generic,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from dateutil.relativedelta import relativedelta
from urllib3.connectionpool import HTTPConnectionPool

from odoo.api import Environment
from odoo.fields import Field
from odoo.sql_db import Cursor

# odoo-typing-classes-generator: imports-insertion-point

T = TypeVar("T")

Domain = List[Union[str, tuple]]


class AbstractModel(Generic[T]):
    create_date: datetime
    env: Environment
    id: int
    ids: List[int]
    write_date: datetime

    @classmethod
    def _browse(
        cls, env: Environment, ids: Tuple[int], prefetch_ids: Collection[int]
    ) -> T:
        pass

    @classmethod
    def _build_model(cls, pool: HTTPConnectionPool, cr: Cursor):
        pass

    @classmethod
    def _build_model_attributes(cls, pool: HTTPConnectionPool):
        pass

    @classmethod
    def _build_model_check_base(cls):
        pass

    @classmethod
    def _build_model_check_parent(model_class, cls, parent_class):  # noqa: N804
        pass

    @classmethod
    def _init_constraints_onchanges(cls, foo: Any = None):
        pass

    @classmethod
    def _patch_method(cls, name: str, method: Callable):
        pass

    @classmethod
    def _revert_method(cls, name: str):
        pass

    @classmethod
    def clear_caches(cls):
        pass

    @classmethod
    def is_transient(cls):
        pass

    def __add__(self, other: T) -> T:
        pass

    def __and__(self, other: T) -> T:
        pass

    def __bool__(self) -> bool:
        pass

    def __contains__(self, item: T) -> bool:
        pass

    def __eq__(self, other: T) -> bool:
        pass

    def __ge__(self, other: T) -> bool:
        pass

    def __getitem__(self, key: Any) -> T:
        pass

    def __gt__(self, other: T) -> bool:
        pass

    def __hash__(self) -> int:
        pass

    def __int__(self) -> int:
        pass

    def __init__(self, pool: HTTPConnectionPool, cr: Cursor):
        pass

    def __iter__(self) -> Iterator[T]:
        pass

    def __le__(self, other: T) -> bool:
        pass

    def __len__(self) -> int:
        pass

    def __lt__(self, other: T) -> bool:
        pass

    def __new__(cls) -> None:
        pass

    def __nonzero__(self) -> bool:
        pass

    def __or__(self, other: T) -> bool:
        pass

    def __repr__(self) -> str:
        pass

    def __setitem__(self, key: Any, value: Any):
        pass

    def __sub__(self, other: T) -> T:
        pass

    def _add_fake_fields(self, fields: dict) -> dict:
        pass

    def _add_field(self, name: str, field: Field):
        pass

    def _add_inherited_fields(self) -> None:
        pass

    def _add_magic_fields(self) -> None:
        pass

    def _add_missing_default_values(self, values: dict) -> dict:
        pass

    def _add_sql_constraints(self) -> None:
        pass

    def _apply_ir_rules(self, query, mode: str = "read"):
        pass

    def _auto_init(self) -> None:
        pass

    def _check_company(self, fnames: Optional[List[str]] = None) -> None:
        pass

    def _check_concurrency(self) -> None:
        pass

    def _check_m2m_recursion(self, field_name: str) -> bool:
        pass

    def _check_qorder(self, word: str) -> bool:
        pass

    def _check_recursion(self, parent: Optional[str] = None) -> bool:
        pass

    def _check_removed_columns(self, log: bool = False):
        pass

    def _compute_display_name(self) -> None:
        pass

    def _compute_field_value(self, field: Field):
        pass

    def _convert_records(
        self, records: T, log: Callable[[T], None] = lambda a: None
    ) -> List[Tuple[Optional[int], Optional[str], dict]]:
        pass

    def _convert_to_record(self, values: dict) -> dict:
        pass

    def _convert_to_write(self, values: dict) -> dict:
        pass

    def _create(self, data_list: List[dict]) -> T:
        pass

    def _create_parent_columns(self) -> None:
        pass

    def _dependent_fields(self, field: Field) -> Iterator[Field]:
        pass

    def _execute_sql(self) -> None:
        pass

    def _export_rows(
        self, fields: List[List[str]], *, _is_toplevel_call: bool = True
    ) -> List[list]:
        pass

    def _extract_records(
        self,
        fields_: List[List[str]],
        data: List[Any],
        log: Callable[[T], None] = lambda a: None,
        limit: Union[int, float] = float("inf"),
    ):
        pass

    def _fetch_field(self, field: Field) -> None:
        pass

    def _fields_view_get(
        self,
        view_id: Optional[int] = None,
        view_type: str = "form",
        toolbar: bool = False,
        submenu: bool = False,
    ) -> dict:
        pass

    def _filter_access_rules(self, operation: str) -> T:
        pass

    def _filter_access_rules_python(self, operation: str) -> T:
        pass

    def _flush_search(
        self,
        domain: Domain,
        fields: Optional[List[str]] = None,
        order: Optional[str] = None,
        seen: Optional[Set[str]] = None,
    ) -> None:
        pass

    # TODO
    def _generate_m2o_order_by(
        self, alias, order_field, query, reverse_direction, seen
    ):
        pass

    # TODO
    def _generate_order_by(self, order_spec, query):
        pass

    # TODO
    def _generate_order_by_inner(
        self, alias, order_spec, query, reverse_direction=False, seen=None
    ):
        pass

    # TODO
    def _generate_translated_field(self, table_alias, field, query):
        pass

    # TODO
    def _get_default_calendar_view(self):
        pass

    # TODO
    def _get_default_form_view(self):
        pass

    # TODO
    def _get_default_graph_view(self):
        pass

    # TODO
    def _get_default_kanban_view(self):
        pass

    # TODO
    def _get_default_pivot_view(self):
        pass

    # TODO
    def _get_default_search_view(self):
        pass

    # TODO
    def _get_default_tree_view(self):
        pass

    # TODO
    def _get_external_ids(self):
        pass

    # TODO
    def _get_placeholder_filename(self, field=None):
        pass

    # TODO
    def _get_xml_ids(self):
        pass

    # TODO
    def _has_onchange(self, field, other_fields):
        pass

    # TODO
    def _in_cache_without(self, field, limit=1000):
        pass

    # TODO
    def _inherits_check(self):
        pass

    # TODO
    def _inherits_join_add(self, current_model, parent_model_name, query):
        pass

    # TODO
    def _inherits_join_calc(self, alias, fname, query):
        pass

    # TODO
    def _init_column(self, column_name):
        pass

    # TODO
    def _is_an_ordinary_table(self):
        pass

    # TODO
    def _load_records(self, data_list, update=False):
        pass

    # TODO
    def _load_records_create(self, values):
        pass

    # TODO
    def _load_records_write(self, values):
        pass

    # TODO
    def _mapped_cache(self, name_seq):
        pass

    # TODO
    def _mapped_func(self, func):
        pass

    # TODO
    def _modified_triggers(self, tree, create=False):
        pass

    # TODO
    def _name_search(
        self, name="", args=None, operator="ilike", limit=100, name_get_uid=None
    ):
        pass

    # TODO
    def _onchange_eval(self, field_name, onchange, result):
        pass

    # TODO
    def _onchange_spec(self, view_info=None):
        pass

    # TODO
    def _parent_store_compute(self):
        pass

    # TODO
    def _parent_store_create(self):
        pass

    # TODO
    def _parent_store_update(self):
        pass

    # TODO
    def _parent_store_update_prepare(self, vals):
        pass

    # TODO
    def _pop_field(self, name):
        pass

    # TODO
    def _populate(self, size):
        pass

    # TODO
    def _populate_factories(self):
        pass

    # TODO
    def _prepare_setup(self):
        pass

    # TODO
    def _read(self, fields):
        pass

    # TODO
    def _read_format(self, fnames, load="_classic_read"):
        pass

    # TODO
    def _read_group_expand_full(self, groups, domain, order):
        pass

    # TODO
    def _read_group_fill_results(
        self,
        domain,
        groupby,
        remaining_groupbys,
        aggregated_fields,
        count_field,
        read_group_result,
        read_group_order=None,
    ):
        pass

    # TODO
    def _read_group_fill_temporal(
        self,
        data,
        groupby,
        aggregated_fields,
        annotated_groupbys,
        interval=relativedelta(months=1),  # noqa: B008
    ):
        pass

    # TODO
    def _read_group_format_result(self, data, annotated_groupbys, groupby, domain):
        pass

    # TODO
    def _read_group_prepare(
        self, orderby, aggregated_fields, annotated_groupbys, query
    ):
        pass

    # TODO
    def _read_group_prepare_data(self, key, value, groupby_dict):
        pass

    # TODO
    def _read_group_process_groupby(self, gb, query):
        pass

    # TODO
    def _read_group_raw(
        self, domain, fields, groupby, offset=0, limit=None, orderby=False, lazy=True
    ):
        pass

    # TODO
    def _read_group_resolve_many2one_fields(self, data, fields):
        pass

    # TODO
    def _rec_name_fallback(self):
        pass

    # TODO
    def _register_hook(self):
        pass

    # TODO
    def _search(
        self,
        args,
        offset=0,
        limit=None,
        order=None,
        count=False,
        access_rights_uid=None,
    ):
        pass

    # TODO
    def _setup_base(self):
        pass

    # TODO
    def _setup_complete(self):
        pass

    # TODO
    def _setup_fields(self):
        pass

    # TODO
    def _table_has_rows(self):
        pass

    # TODO
    def _unregister_hook(self):
        pass

    # TODO
    def _update_cache(self, values, validate=True):
        pass

    # TODO
    def _valid_field_parameter(self, field, name):
        pass

    # TODO
    def _validate_fields(self, field_names, excluded_names=()):
        pass

    # TODO
    def _where_calc(self, domain, active_test=True):
        pass

    # TODO
    def _write(self, vals):
        pass

    # TODO
    def action_archive(self):
        pass

    # TODO
    def action_unarchive(self):
        pass

    def browse(self, ids: Union[int, List[int]]) -> T:
        pass

    # TODO
    def check_access_rights(self, operation, raise_exception=True):
        pass

    # TODO
    def check_access_rule(self, operation):
        pass

    # TODO
    def check_field_access_rights(self, operation, fields):
        pass

    # TODO
    def compute_concurrency_field(self):
        pass

    # TODO
    def compute_concurrency_field_with_access(self):
        pass

    # TODO
    def concat(self, *args):
        pass

    def copy(  # pylint: disable=method-required-super
        self, default: Optional[dict] = None
    ) -> T:
        pass

    # TODO
    def copy_data(self, default=None):
        pass

    def copy_translations(
        old,  # noqa: N805
        new,
        excluded=(),
    ):
        pass

    def create(self, vals_list: Union[List[dict], dict]) -> T:  # pylint: disable=method-required-super
        pass

    # TODO
    def default_get(  # pylint: disable=method-required-super
        self, fields_list
    ):
        pass

    # TODO
    def ensure_one(self):
        pass

    def exists(self) -> T:
        pass

    # TODO
    def export_data(self, fields_to_export):
        pass

    # TODO
    def fields_get(self, allfields=None, attributes=None):
        pass

    # TODO
    def fields_get_keys(self):
        pass

    # TODO
    def fields_view_get(
        self,
        view_id=None,
        view_type="form",
        toolbar=False,
        submenu: Union["Model", bool] = False,
    ):
        pass

    def filtered(self, func: Union[Callable[[T], bool], str]) -> T:
        pass

    # TODO
    def filtered_domain(self, domain):
        pass

    # TODO
    def flush(self, fnames=None, records=None):
        pass

    # TODO
    def get_access_action(self, access_uid=None):
        pass

    # TODO
    def get_base_url(self):
        pass

    # TODO
    def get_empty_list_help(self, help):
        pass

    # TODO
    def get_external_id(self):
        pass

    # TODO
    def get_formview_action(self, access_uid=None):
        pass

    # TODO
    def get_formview_id(self, access_uid=None):
        pass

    # TODO
    def get_metadata(self):
        pass

    # TODO
    def get_xml_id(self):
        pass

    # TODO
    def init(self):
        pass

    # TODO
    def invalidate_cache(self, fnames=None, ids=None):
        pass

    # TODO
    def load(self, fields, data):
        pass

    # TODO
    def load_views(self, views, options=None):
        pass

    # TODO
    def mapped(self, func):
        pass

    # TODO
    def modified(self, fnames, create=False, before=False):
        pass

    def name_create(self, name: str) -> T:
        pass

    def name_get(self) -> List[Tuple[int, str]]:
        pass

    def name_search(
        self,
        name: str = "",
        args: Optional[Domain] = None,
        operator: str = "ilike",
        limit: int = 100,
    ) -> List[Tuple[int, str]]:
        pass

    def new(  # pylint: disable=dangerous-default-value
        self,
        values: dict = {},  # noqa: B006
        origin: Optional["Model"] = None,
        ref: Optional[str] = None,
    ):
        pass

    def onchange(
        self, values: dict, field_name: Union[str, List[str]], field_onchange: dict
    ) -> Optional[dict]:
        pass

    def read(  # pylint: disable=method-required-super
        self, fields: Optional[List[str]] = None, load: str = "_classic_read"
    ) -> List[dict]:
        pass

    def read_group(
        self,
        domain: Domain,
        fields: List[str],
        groupby: List[str],
        offset: int = 0,
        limit: Optional[int] = None,
        orderby: Optional[str] = False,
        lazy: bool = True,
    ) -> List[dict]:
        pass

    def recompute(
        self, fnames: Optional[List[str]] = None, records: Optional["Model"] = None
    ) -> None:
        pass

    def refresh(self) -> None:
        pass

    def search(
        self,
        args: Domain,
        offset: int = 0,
        limit: Optional[int] = None,
        order: Optional[str] = None,
        count: bool = False,
    ) -> T:
        pass

    def search_count(self, args: Domain) -> int:
        pass

    def search_read(
        self,
        domain: Optional[Domain] = None,
        fields: Union[List[str], tuple, None] = None,
        offset: int = 0,
        limit: Optional[int] = None,
        order: Optional[str] = None,
    ) -> List[dict]:
        pass

    def sorted(self, key: Optional[Union[Callable[[T], Any], str]], reverse: bool) -> T:
        pass

    def sudo(self, flag: bool = True) -> T:
        pass

    def toggle_active(self) -> None:
        pass

    def union(  # pylint: disable=method-required-super
        self, *args: "Model"
    ) -> None:
        pass

    def unlink(  # pylint: disable=method-required-super
        self,
    ) -> bool:
        pass

    def update(self, values: dict) -> None:
        pass

    def user_has_groups(self, groups: str) -> bool:
        pass

    def view_header_get(
        self, view_id=None, view_type: str = "form"
    ) -> Union[str, bool]:
        pass

    def view_init(self, fields_list: List[str]) -> None:
        pass

    def write(  # pylint: disable=method-required-super
        self, vals: dict
    ) -> bool:
        pass

    def with_company(
        self,
        company: "ResCompany",  # noqa: F821
    ) -> T:
        pass

    def with_context(self, *args, **kwargs) -> T:
        pass

    def with_env(self, env: Environment) -> T:
        pass

    def with_prefetch(self, prefetch_ids: Optional[List[int]] = None) -> T:
        pass

    def with_user(
        self,
        user: "ResUsers",  # noqa: F821
    ) -> T:
        pass


class Model(AbstractModel[T]):
    pass


class TransientModel(Model[T]):
    # TODO
    def _transient_clean_old_rows(self, max_count):
        pass

    # TODO
    def _transient_clean_rows_older_than(self, seconds):
        pass

    # TODO
    def _transient_vacuum(self):
        pass


# odoo-typing-classes-generator: classes-insertion-point



import logging

import click

from odoo_typing_classes_generator.core.generator import Generator

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
_logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "--modules",
    help="Comma-separated list of Odoo modules to generate typing classes for.",
    required=True,
)
@click.option(
    "--addons-path",
    required=True,
    help="Path where the modules are located, relative to the current working directory.",
)
@click.option(
    "--generate-all-classes",
    is_flag=True,
    help="If set, the models.py files will be filled with all the possible Odoo models. "
    "If not set, the models.py files will only be created if they don't exist yet "
    "and won't be modified if they already exist.",
)
def main(modules: str, addons_path: str, generate_all_classes: bool):
    for module in modules.split(","):
        _logger.info(f"[{module}] Generating typing classes...")
        Generator(addons_path, generate_all_classes).generate(module)
        _logger.info(f"[{module}] Done generating typing classes.")


if __name__ == "__main__":
    main()



from typing import List

from odoo import fields, models

from odoo_typing_classes_generator.tests.fake_module_1 import typing as models_typing


class FakeModel1(models.Model):
    _name = "fake.model.1"
    _inherit = ["res.partner"]
    _module = ""

    a_text_field = fields.Text()
    a_related_field = fields.Char(related="many_to_one_field_id.name")
    a_many_to_one_field = fields.Many2one("res.company")

    def a_documented_function(self, _an_argument: int = 1) -> models_typing.ResCountry:
        return self.a_many_to_one_field.country_id

    def an_undocumented_method(self, _an_argument=""):
        return

    @staticmethod
    def a_static_method_with_default_arguments(
        an_argument=lambda: 0, another_argument="world"
    ):
        return f"[{an_argument()}] Hello {another_argument}!"

    @staticmethod
    def a_static_method(an_argument: bool) -> int:
        return 42 if an_argument else 0

    @classmethod
    def a_class_method(cls, an_argument: str) -> List[str]:
        return [cls._name, an_argument]



import dataclasses
import inspect
import keyword
import logging
import re
from collections import defaultdict
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    get_args,
    get_origin,
)

from odoo.tools import ConstantMapping

from ..assets import template as models_typing

_logger = logging.getLogger(__name__)


def _get_field_names_to_ignore_by_class() -> Dict[type, Set[str]]:
    to_ignore_by_class = defaultdict(set)
    for module in [
        models_typing.AbstractModel,
        models_typing.Model,
        models_typing.TransientModel,
    ]:
        to_ignore_by_class[module] |= set(module.__annotations__.keys())
    # Inherit the fields from parent classes
    to_ignore_by_class[models_typing.Model] |= to_ignore_by_class[
        models_typing.AbstractModel
    ]
    to_ignore_by_class[models_typing.TransientModel] |= to_ignore_by_class[
        models_typing.Model
    ]
    return dict(to_ignore_by_class)


def _get_function_names_to_ignore_by_class() -> Dict[type, Set[str]]:
    to_ignore_by_class = defaultdict(set)
    for module in [
        models_typing.AbstractModel,
        models_typing.Model,
        models_typing.TransientModel,
    ]:
        to_ignore_by_class[module] |= {
            function_name
            for function_name, _function in inspect.getmembers(
                module,
                lambda member: (inspect.isfunction(member) or inspect.ismethod(member)),
            )
        }
    # Inherit the functions from parent classes
    to_ignore_by_class[models_typing.Model] |= to_ignore_by_class[
        models_typing.AbstractModel
    ]
    to_ignore_by_class[models_typing.TransientModel] |= to_ignore_by_class[
        models_typing.Model
    ]
    return dict(to_ignore_by_class)


field_names_to_ignore_by_class: Dict[
    type, Set[str]
] = _get_field_names_to_ignore_by_class()
function_names_to_ignore_by_class: Dict[
    type, Set[str]
] = _get_function_names_to_ignore_by_class()


def _snake_case_to_pascal_case(snake_case_str: str) -> str:
    """
    Converts a snake_case string to PascalCase.
    """
    pascal_case_str = ""
    for word in snake_case_str.split("_"):
        if not word:
            continue
        pascal_case_str += word[0].upper()
        if len(word) > 1:
            pascal_case_str += word[1:]
    return pascal_case_str


@dataclasses.dataclass
class ImportData:
    package_name: str
    class_name: Optional[str] = None
    alias: Optional[str] = None

    def serialize(self) -> str:
        if self.class_name:
            import_str = f"from {self.package_name} import {self.class_name}"
        else:
            import_str = f"import {self.package_name}"
        if self.alias:
            import_str += f" as {self.alias}"
        return import_str

    def __hash__(self) -> int:
        return hash((self.package_name, self.class_name, self.alias))


@dataclasses.dataclass
class TypeData:
    value: type
    imports: Set[ImportData] = dataclasses.field(init=False)

    def __post_init__(self):
        self.imports = TypeData.collect_imports(self.value)

    @staticmethod
    def collect_imports(type_hint: type) -> Set[ImportData]:
        if type_hint.__module__ == "builtins":
            return set()
        if type_hint.__module__ != "typing":
            return {
                ImportData(
                    package_name=type_hint.__module__,
                    class_name=type_hint.__name__,
                ),
            }
        origin_type = get_origin(type_hint)
        if origin_type is None:
            imports = {
                ImportData(
                    package_name="typing",
                    class_name="Any",
                ),
            }
        elif origin_type.__module__ == "typing":
            imports = {
                ImportData(
                    package_name=origin_type.__module__,
                    class_name=origin_type._name,
                ),
            }
        elif origin_type.__module__ == "builtins":
            imports = {
                ImportData(
                    package_name="typing",
                    class_name=origin_type.__name__.capitalize(),
                ),
            }
        elif origin_type.__name__ in ["Callable", "Generator"]:
            imports = {
                ImportData(
                    package_name="typing",
                    class_name=origin_type.__name__,
                ),
            }
        else:
            imports = {
                ImportData(
                    package_name=origin_type.__module__,
                    class_name=origin_type.__name__,
                ),
            }
        for arg in get_args(type_hint):
            if not isinstance(arg, list):
                imports |= TypeData.collect_imports(arg)
                continue
            for item in arg:
                imports |= TypeData.collect_imports(item)
        return imports

    def serialize(self) -> str:
        if self.value == type(None):
            return "None"
        if self.value.__module__ != "typing":
            return self.value.__name__
        value_str = str(self.value)
        for import_data in self.imports:
            value_str = value_str.replace(f"{import_data.package_name}.", "")
        return value_str.replace("NoneType", "None")


@dataclasses.dataclass
class FunctionArgumentData:
    name: str
    type: Optional[TypeData] = None
    default: Any = None
    has_default: bool = False
    is_vararg: bool = False
    is_kwarg: bool = False

    def serialize(self) -> str:
        arg_str = ""
        if self.is_vararg:
            arg_str += "*"
        elif self.is_kwarg:
            arg_str += "**"
        arg_str += self.name
        if self.type:
            arg_str += f": {self.type.serialize()}"
        if self.has_default:
            if self.type:
                arg_str += " = "
            else:
                arg_str += "="
            if self.default is None or self.default == type(None):
                arg_str += "None"
            elif type(self.default) == type:
                arg_str += self.default.__name__
            elif isinstance(self.default, str):
                # Escape quotes in string defaults
                default = self.default.replace('"', '\\"')
                arg_str += f'"{default}"'
            elif isinstance(self.default, ConstantMapping):
                arg_str += f"ConstantMapping({self.default[0]})"
            else:
                arg_str += str(self.default)
        return arg_str


@dataclasses.dataclass
class FunctionData:
    name: str
    args: List[FunctionArgumentData]
    type: Union[TypeData, str, None] = None
    is_class_method: bool = False
    is_static_method: bool = False

    def serialize(self, escape_pattern: re.Pattern) -> str:
        function_str = ""
        if self.is_class_method:
            function_str += "    @classmethod\n"
        elif self.is_static_method:
            function_str += "    @staticmethod\n"
        function_str += f"    def {self.name}("
        function_str += ", ".join(arg.serialize() for arg in self.args)
        function_str += ")"
        if self.type:
            function_str += f" -> {self.type.serialize()}"
        function_str = re.sub(escape_pattern, r'"\1"', function_str)
        function_str += ":\n        pass\n"
        return function_str


@dataclasses.dataclass
class FieldData:
    name: str
    type: Union[TypeData, str]
    related: Optional[str]

    def serialize(
        self, merged_models_data_by_odoo_model_name: Dict[str, "MergedModelData"]
    ) -> str:
        if isinstance(self.type, str):
            serialized_type = merged_models_data_by_odoo_model_name[
                self.type
            ].unique_class_name
            # We cannot use Literal[False] here, because when iterating over the field,
            # the IDE wouldn't know which of the two types the elements have.
            serialized_type = f'Union["{serialized_type}", bool]'
        else:
            serialized_type = self.type.serialize()
        field_definition = f"{self.name}: {serialized_type}"
        if keyword.iskeyword(self.name):
            field_definition = (
                f"# {field_definition}  # This field name is a Python keyword"
            )
        return f"    {field_definition}"

    def __hash__(self) -> int:
        return hash((self.name, self.type, self.related))


@dataclasses.dataclass
class ModelData:
    """
    Represents a model dependency, this information will be used to generate
    the list of imports and inherited classes for a given model's stub
    """

    package_name: str  # Example "odoo.addons.stock.models.stock_move_line"
    module_name: str = dataclasses.field(init=False)  # Example "stock_move_line"
    odoo_model_name: str  # Example "stock.move.line"
    class_name: str = dataclasses.field(init=False)  # Example "StockMoveLine"
    source_class_name: str  # Example "StockMoveLine"
    addon_name: str  # Example "stock"
    unique_class_name: str = dataclasses.field(
        init=False
    )  # Example "StockStockMoveLine"
    inherited_model_names: Set[str]  # Example {}
    field_data_by_name: Dict[
        str,
        FieldData,
    ]  # Example {"picking_id": "stock.picking", ..., "product_qty": float, ...,  "date": datetime.datetime, ... }
    function_data_by_name: Dict[
        str,
        FunctionData,
    ]  # Example {"action_assign": FunctionData(name="action_assign", return_type=None, unknown_return_type=False, args=[FunctionArgumentData(name="self", type=None, unknown_type=True, default=None, has_default=False, is_vararg=False, is_kwarg=False)])}
    is_base_definition: bool = dataclasses.field(init=False)  # Example True
    # All the imports that are needed to define this model
    imports: Set[ImportData]
    abstract: bool
    transient: bool

    def __post_init__(self):
        """
        Assigns a unique class name based on the addon name and the model name.

        For example, the model 'stock.move.line' from the 'stock' addon has a unique
        class name 'StockStockMoveLine'.

        This enables to manipulate different implementations of the same model.
        """
        self.class_name = self.source_class_name
        self.module_name = self.package_name.rsplit(".", 1)[1]
        self.unique_class_name = _snake_case_to_pascal_case(
            self.odoo_model_name.replace(".", "_")
        )
        self.is_base_definition = self.odoo_model_name not in self.inherited_model_names

    def get_class_definition(self) -> str:
        return f"class {self.unique_class_name}:\n    pass\n"

    def get_stub_definition(
        self,
        merged_models_data_by_odoo_model_name: Dict[str, "MergedModelData"],
    ) -> Tuple[Set[str], str]:
        import_classes_to_ignore = {
            model.unique_class_name
            for model in merged_models_data_by_odoo_model_name.values()
        }
        import_lines: Set[str] = {
            import_data.serialize()
            for import_data in self.imports
            if import_data.class_name not in import_classes_to_ignore
        }
        if self.transient:
            base_class = models_typing.TransientModel
        elif self.abstract:
            base_class = models_typing.AbstractModel
        else:
            base_class = models_typing.Model
        field_names_to_ignore = set(field_names_to_ignore_by_class[base_class])
        function_names_to_ignore = set(function_names_to_ignore_by_class[base_class])
        inherited_classes = [
            f'"{self.unique_class_name}"',
        ]
        for inherited_model_name in sorted(
            self.inherited_model_names - {self.odoo_model_name}
        ):
            inherited_model_data = merged_models_data_by_odoo_model_name[
                inherited_model_name
            ]
            inherited_classes.append(f'"{inherited_model_data.unique_class_name}"')
            for field_name in inherited_model_data.field_data_by_name:
                field_names_to_ignore.add(field_name)
            for function_name in inherited_model_data.function_data_by_name:
                function_names_to_ignore.add(function_name)
        class_definition = f"class {self.unique_class_name}({base_class.__name__}["
        if len(inherited_classes) == 1:
            class_definition += inherited_classes[0]
        else:
            class_definition += f"Union[{', '.join(inherited_classes)}]"
        class_definition += "]):\n"
        class_definition += self._get_class_documentation()
        class_definition += f'\n    _name = "{self.odoo_model_name}"\n'
        field_data_by_name = {
            field_name: field_data
            for field_name, field_data in self.field_data_by_name.items()
            if field_name not in field_names_to_ignore
        }
        function_data_by_name = {
            function_name: function_data
            for function_name, function_data in self.function_data_by_name.items()
            if function_name not in function_names_to_ignore
        }
        if not field_data_by_name and not function_data_by_name:
            return import_lines, class_definition
        if field_data_by_name:
            field_lines = {
                field_data.serialize(merged_models_data_by_odoo_model_name)
                for field_data in field_data_by_name.values()
            }
            class_definition += "\n" + "\n".join(sorted(field_lines)) + "\n"
        merged_models_class_names = {
            merged_models_data.unique_class_name
            for merged_models_data in merged_models_data_by_odoo_model_name.values()
        }
        escape_pattern = rf"\b({'|'.join(merged_models_class_names)})\b"
        if function_data_by_name:
            function_lines = {
                function_data.serialize(escape_pattern)
                for function_data in function_data_by_name.values()
            }
            class_definition += "\n" + "\n".join(sorted(function_lines))
        return import_lines, class_definition

    def _get_class_documentation(self) -> str:
        return ""

    def __hash__(self) -> int:
        return hash(self.unique_class_name)


@dataclasses.dataclass
class MergedModelData(ModelData):
    # All the models that were merged to create this model
    source_models: List[ModelData]

    def _get_class_documentation(self) -> str:
        documentation = '    """\n'
        documentation += f"    Merged model for {self.odoo_model_name}, built from:\n"
        source_model_names = [
            f"{source_model.package_name}.{source_model.source_class_name}"
            for source_model in self.source_models
        ]
        for source_model_name in sorted(source_model_names):
            documentation += f"        * {source_model_name}\n"
        inherited_model_names = sorted(
            self.inherited_model_names - {self.odoo_model_name}
        )
        if inherited_model_names:
            documentation += "\n    And inherits from:\n"
            for inherited_model_name in inherited_model_names:
                documentation += f"        * {inherited_model_name}\n"
        documentation += '    """\n'
        return documentation


@dataclasses.dataclass
class AddonData:
    name: str
    models_data_by_odoo_model_name: Dict[
        str,
        List[ModelData],
    ]  # All the models defined or updated in the addon, there might be some inheritance for within one module
    dependencies: Set["AddonData"]  # All the dependencies

    def __hash__(self) -> int:
        return hash(self.name)
